#include <qcolor.h>
#include <qevent.h>
#include <qimage.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/Configuration/src/Configuration.h"
#include "../../../Core/src/DefaultMouseHandler.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../DataTypes/src/MyDataTypes.h"
#include "EWAGLRendererConfiguration.h"
#include "EWAGLRendererWidget.h"

#include <iostream>
#include <cmath>

#define DEFAULT_CUTOFF_RADIUS 1
#define DEFAULT_SPLAT_RADIUS 1
#define DEFAULT_MAX_OVERLAP 8
#define DEFAULT_DEPTH_OFFSET 1
#define GAUSS_TEXTURE_SIZE   256
#define CG_PROGRAM_FILENAME "Renderers/CG/hwpointrenderer.cg"


// **************
// public methods
// **************

EWAGLRendererWidget::EWAGLRendererWidget (const bool isVisible, EWAGLRendererConfiguration *rendererConfiguration,
											QWidget *parent, const char *name)
  : QGLWidget(QGLFormat(Rgba | DepthBuffer | AlphaChannel | DoubleBuffer), parent, name),
    modelInteractState(RendererInterface::INTERACT_STATIC),
    viewpointInteractState(RendererInterface::INTERACT_STATIC)
{
  Scene         *scene;
  Configuration *configuration;

  this->rendererConfiguration = rendererConfiguration;

  // disable clearing the background before each paint event
  this->setBackgroundMode (NoBackground);

  updateViewPortSize = true;
  shadingEnabled     = true;
  visible            = isVisible;

  {
    configuration = Configuration::getInstance();
    viewFrustum = configuration->getViewFrustum();
  }

  MtrUnity4x4f (sceneViewMatrix);

  scene = Scene::getInstance();

  // get informed whenever the Objects in the Scene change
  this->connect (scene, SIGNAL (objectModified(bool)),
				 this, SLOT (handleObjectModified()));

  // get informed whenever the active Object in the Scene changes
  this->connect (scene, SIGNAL (activeObjectChanged()),
				 this, SLOT (handleActiveObjectChanged()));

  // get informed whenever Objects are added to the Scene
  this->connect (scene, SIGNAL (objectAdded()),
				 this, SLOT (handleObjectAdded()));

  // get informed whenever Objects are added to the Scene
  this->connect (scene, SIGNAL (objectRemoved()),
				 this, SLOT (handleObjectRemoved()));

  this->connect (rendererConfiguration, SIGNAL (cutOffRadiusChanged (const float)),
				 this, SLOT (handleCutOffRadiusChanged (const float)));

  this->connect (rendererConfiguration, SIGNAL (gaussDecayRadiusChanged (const float)),
				 this, SLOT (handleGaussDecayRadiusChanged (const float)));
  
  this->connect (CoreTools::getInstance()->getSelectionTool(), SIGNAL (selectionChanged (SurfelInterface::Flag)),
		           this, SLOT (handleSelectionChanged (SurfelInterface::Flag)));
	
  // get informed whenever all the settings in the EWAGLRendererConfiguration have been applied
  this->connect (rendererConfiguration, SIGNAL (configurationApplied()),
				 this, SLOT (handleConfigurationApplied()));

  bShowSelection = true;
  
  backgroundColor = 0xFF8598b5;

}

EWAGLRendererWidget::~EWAGLRendererWidget() {
}

void EWAGLRendererWidget::setVisible (const bool enable) {
	visible = enable;
}

bool EWAGLRendererWidget::isVisible() const {
	return visible;
}

QSize EWAGLRendererWidget::sizeHint() const {
	return Configuration::getInstance()->getViewPortSize();
}

QSize EWAGLRendererWidget::minimumSizeHint() const {
	return Configuration::getInstance()->getMinimumViewPortSize();
}

QSizePolicy EWAGLRendererWidget::sizePolicy() const {
	return QSizePolicy (QSizePolicy::Preferred , QSizePolicy::Preferred);
}

void EWAGLRendererWidget::setShadingEnabled (const bool enable) {

	this->makeCurrent();

	shadingEnabled = enable;

	// only make changes if enable is different from the previous
	// settings, since we need to recompile the display lists
	if ((enable == true) && (glIsEnabled (GL_LIGHTING) == GL_FALSE)) {
		glEnable (GL_LIGHTING);
		glEnable (GL_LIGHT0);
	}
	else if ((enable == false) && glIsEnabled (GL_LIGHTING) == GL_TRUE) {
		glDisable (GL_LIGHTING);
		glDisable (GL_LIGHT0);
	}

}

bool EWAGLRendererWidget::isShadingEnabled() const {
	return glIsEnabled (GL_LIGHTING);
}

void EWAGLRendererWidget::setLightDirection (const Vector3D newLightDirection) {
	lightDirection = newLightDirection;
	this->updateLightDirection (true);
}

Vector3D EWAGLRendererWidget::getLightDirection() const {
	return lightDirection;
}


void EWAGLRendererWidget::setBackgroundColor(QRgb newBackgroundColor) {
	backgroundColor = QColor(newBackgroundColor);
	glClearColor(backgroundColor.red() / 255.0f, backgroundColor.green() / 255.0f, backgroundColor.blue() / 255.0f, 1.0);
}

QRgb EWAGLRendererWidget::getBackgroundColor() const {
	return backgroundColor.rgb();
}

void EWAGLRendererWidget::setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix) {
	memcpy (sceneViewMatrix, newSceneViewMatrix, sizeof (MyDataTypes::TransformationMatrix16f));
}

void EWAGLRendererWidget::getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const {
	memcpy (sceneViewMatrix, this->sceneViewMatrix, sizeof (MyDataTypes::TransformationMatrix16f));
}

void EWAGLRendererWidget::setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum) {

	viewFrustum = newViewFrustum;

	this->makeCurrent();
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();
	gluPerspective (viewFrustum.fieldOfView, viewFrustum.aspectRatio,
		            viewFrustum.nearPlane,   viewFrustum.farPlane);
	glMatrixMode (GL_MODELVIEW);

}

MyDataTypes::ViewFrustum EWAGLRendererWidget::getViewFrustum() const {
	return viewFrustum;
}

void EWAGLRendererWidget::setViewPortSize (const QSize newViewPortSize) {
	// we do not want to update the configuration, it has already been updated
	// see resizeGL()
	updateViewPortSize = false;
	this->resize (newViewPortSize);
	// from now on we want to update the view port size in the Configuration again,
	// that is whenever this widget is resized by the user
	updateViewPortSize = true;
}
QSize EWAGLRendererWidget::getViewPortSize() const {
	return this->size();
}

void EWAGLRendererWidget::setSelectionVisualizationEnabled(const bool showSelection) {

	if (bShowSelection != showSelection) {
		bShowSelection = showSelection;
		buildVertexArrays();	
	}
}

void EWAGLRendererWidget::handleSelectionChanged (SurfelInterface::Flag) {
	buildVertexArrays();	
}


bool EWAGLRendererWidget::isSelectionVisualizationEnabled() const {
	return bShowSelection;
}

// *****************
// protected methods
// *****************

void EWAGLRendererWidget::initializeGL() {

#ifdef _WIN32	
	init_GL_extensions();
#endif

    firstCall = true;

	lightDirection = Configuration::getInstance()->getLightDirection();
	
    this->updateLightDirection (false);

    this->setShadingEnabled (shadingEnabled);
  
	QRgb backgroundColor = Configuration::getInstance()->getBackgroundColor();
	setBackgroundColor(backgroundColor);

	// set the view frustum
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();
	gluPerspective (viewFrustum.fieldOfView, viewFrustum.aspectRatio,
		            viewFrustum.nearPlane,   viewFrustum.farPlane);
	glMatrixMode (GL_MODELVIEW);
	
    handleCutOffRadiusChanged(rendererConfiguration->getCutOffRadius());
    handleGaussDecayRadiusChanged(rendererConfiguration->getGaussDecayRadius());

	renderer.init();

}

void EWAGLRendererWidget::paintGL() {
	static const float invSqrt3 = 1.0f / sqrt(3.0f);

    if (firstCall) {
      firstCall = false;
	  handleObjectAdded();
	}

//	glClearColor(0.52f, 0.60f, 0.71f, 1.0f);
	glClearDepth(1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  	Scene* scene = Scene::getInstance();

	int b = 0;

	for (Object* object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {

		if (object->isFlagOn (Object::VISIBLE))  {
	
			MyDataTypes::TransformationMatrix16f objectTransformation, finalTransformation;
			object->getTransformationMatrix (objectTransformation);
			MtrMult4x4f (sceneViewMatrix, objectTransformation, finalTransformation);
			glLoadMatrixf (finalTransformation);

			// calculate depth offset
			Vector3D scale;
			scale[0] = (finalTransformation[0] + finalTransformation[4] + finalTransformation[8]) * invSqrt3;
			scale[1] = (finalTransformation[1] + finalTransformation[5] + finalTransformation[9]) * invSqrt3;
			scale[2] = (finalTransformation[2] + finalTransformation[6] + finalTransformation[10]) * invSqrt3;
			double s = scale.getLength() * 2;
			float depthOffset = averageRadii[b] * s;

			renderer.depthOffset(depthOffset);

			SurfelCollection* surfelCollection = object->getSurfelCollection();

			int nOfSurfels = surfelCollection->getNofSurfels();

			glInterleavedArrays(GL_T2F_C4F_N3F_V3F, 0, (GLvoid*)&(vertexArrays[b])[0]);
			
			renderer.render(0, nOfSurfels);
			
		}

		++b;
	}

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);

	// **********************************************************
	// * Callback - Extension
	glLoadMatrixf(sceneViewMatrix);
	RendererInterface::callPostRenderCBs();

    emit repainted();
}

void EWAGLRendererWidget::resizeGL (int width, int height) {

	Configuration            *configuration;

    glViewport (0, 0, width, height);

    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity();

    {
	  configuration = Configuration::getInstance();
	  if (updateViewPortSize == true && visible == true) {
		  configuration->setViewPortSize (QSize (width, height));
	  }
    }
}

void EWAGLRendererWidget::mousePressEvent (QMouseEvent *me) {	
	DefaultMouseHandler::handleMousePressEvent (me);
}

void EWAGLRendererWidget::mouseMoveEvent (QMouseEvent *me) {
	DefaultMouseHandler::handleMouseMoveEvent (me);
}

void EWAGLRendererWidget::mouseReleaseEvent (QMouseEvent *me) {
	DefaultMouseHandler::handleMouseReleaseEvent (me);
}

void EWAGLRendererWidget::keyPressEvent (QKeyEvent *ke) {
	ke->ignore();
}

void EWAGLRendererWidget::wheelEvent (QWheelEvent *we) {
	DefaultMouseHandler::handleMouseWheelEvent(we);
	
}

// ***************
// private methods
// ***************

void EWAGLRendererWidget::updateLightDirection (const bool updateScene) {

	GLfloat glLightPosition[4];   
	
	// make it the current EWAGL rendering context
	this->makeCurrent();

    // We get a direction, so set the light source far away...
	glLightPosition[0] = -10000.f * lightDirection[0];
	glLightPosition[1] = -10000.f * lightDirection[1];
	glLightPosition[2] = -10000.f * lightDirection[2];
	glLightPosition[3] = 1.0f;

	// put EWAGL light at absolute position 'glLightPosition' (not relative to current model view matrix)
	glPushMatrix();
	glLoadIdentity();
    glLightfv (GL_LIGHT0, GL_POSITION, glLightPosition);
	glPopMatrix();

	if (updateScene == true) {
		this->update();
	}
}

void EWAGLRendererWidget::buildVertexArrays() {
  // get scene
  Scene* scene = Scene::getInstance();

  // build vertex array for each object
  int b = 0;
  double averageRadius = 0;
  long surfelCount = 0;
  float red, green, blue;

  vertexArrays.resize(Scene::getInstance()->getNofObjects());
  averageRadii.resize(Scene::getInstance()->getNofObjects());


  for (Object* object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
	  // get num surfels
	  SurfelCollection* surfelCollection = object->getSurfelCollection();
	  int nOfSurfels = surfelCollection->getNofSurfels();
  
	  // allocate memory
	  std::vector<MySurfel>& vertexArray = vertexArrays[b];
	  vertexArray.resize(nOfSurfels);

	  // copy vertices into buffer
	  int i = 0;
	  for (SurfelInterface* surfel = surfelCollection->getFirstSurfel(); surfel != 0; surfel = surfelCollection->getNextSurfel()) {
		Vector3D position = surfel->getPosition();
		QRgb color = surfel->getDiffuseColor();
		Vector3D t1 = surfel->getAxisOne();
		Vector3D t2 = surfel->getAxisTwo();

		double radius = surfel->getRadius();
		averageRadius += radius;
		surfelCount++;

		red = qRed(color) / 255.0f;	
		blue = qGreen(color) / 255.0;
		green = qBlue(color) / 255.0;

		showSelection(surfel, red, blue, green);

		vertexArray[i].position[0] = position[0];
		vertexArray[i].position[1] = position[1];
		vertexArray[i].position[2] = position[2];
		vertexArray[i].normal[0] = t1[0];
		vertexArray[i].normal[1] = t1[1];
		vertexArray[i].normal[2] = t1[2];	
		vertexArray[i].texture[0] = t2[0];
		vertexArray[i].texture[1] = t2[1];
		vertexArray[i].color[3]   = t2[2];

		vertexArray[i].color[0] = red;
		vertexArray[i].color[1] = blue;
		vertexArray[i].color[2] = green;

		++i;
	  }

	  averageRadius /= surfelCount;
	  averageRadii[b] = averageRadius;

	  ++b;
  }


}

///////////////////////////////////////////////////////////////////////////////

// *************
// private slots
// *************

void EWAGLRendererWidget::handleObjectModified () {
  buildVertexArrays();
}

void EWAGLRendererWidget::handleActiveObjectChanged() {
}

void EWAGLRendererWidget::handleObjectAdded() {
  buildVertexArrays();
}

void EWAGLRendererWidget::handleObjectRemoved() {
  buildVertexArrays();
}

void EWAGLRendererWidget::handleCutOffRadiusChanged (const float newCutOffRadius) {

	renderer.splatRadius(newCutOffRadius);
}

void EWAGLRendererWidget::handleGaussDecayRadiusChanged (const float newGaussRadius) {

	renderer.cutoffRadius(newGaussRadius);
}

void EWAGLRendererWidget::handleConfigurationApplied()
{
  this->update();
}

// interaction stuff //////////////////////////////////////////////////////////

RendererInterface::InteractStateEnum  EWAGLRendererWidget::getModelInteractState(void) const
{
  return modelInteractState;
}

RendererInterface::InteractStateEnum  EWAGLRendererWidget::getViewpointInteractState(void) const
{
  return viewpointInteractState;
}

void  EWAGLRendererWidget::setInteractState(RendererInterface::InteractStateEnum state,
											 RendererInterface::InteractStateAspectEnum target)
{
  if (target & RendererInterface::INTERACT_MODEL)      modelInteractState = state;
  if (target & RendererInterface::INTERACT_VIEWPOINT)  viewpointInteractState = state;
}

QImage EWAGLRendererWidget::renderPixmap() {

    QImage res;
    int w = this->size().width();
    int h = this->size().height();

	res = QImage( w, h, 32 );
	glReadPixels( 0, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, res.bits() );
	if ( QImage::systemByteOrder() == QImage::BigEndian ) {
	    // OpenGL gives RGBA; Qt wants ARGB
	    uint *p = (uint*)res.bits();
	    uint *end = p + w*h;
		while ( p < end )
		    *p++ >>= 8;
	}
    res = res.swapRGB();
	return res.mirror();
}


void errorCallback() {
  std::cerr << "CG error: " << cgGetErrorString(cgGetError()) << std::endl;
}

Renderer::Renderer() :
  cutoff_radius(DEFAULT_CUTOFF_RADIUS),
  splat_radius(DEFAULT_SPLAT_RADIUS),
  max_overlap(DEFAULT_MAX_OVERLAP),
  depth_offset(DEFAULT_DEPTH_OFFSET) {
}

void Renderer::generateGaussTexture() {
  glBindTexture(GL_TEXTURE_1D, gl_gauss_texture);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  float g[GAUSS_TEXTURE_SIZE];
  for (int i = 0; i < GAUSS_TEXTURE_SIZE; i++) {
    double x = i * cutoff_radius * cutoff_radius / 
        double(GAUSS_TEXTURE_SIZE - 1);
    float f = exp(-0.5*x);
    g[i] = f / max_overlap;
  }

  glTexImage1D(GL_TEXTURE_1D, 0, GL_ALPHA, GAUSS_TEXTURE_SIZE, 0, GL_ALPHA, GL_FLOAT, g);
}

void Renderer::init() {
  ///////////////////////
  // Cg Initialization //
  ///////////////////////

  cgSetErrorCallback(errorCallback);

  // Create Context
  cg_context = cgCreateContext();

  // create vertex profile
  cg_vertex_profile = CG_PROFILE_ARBVP1;
  cgGLSetOptimalOptions(cg_vertex_profile);

  // load vertex programs
  cg_vertex_program_1 = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_vertex_profile, "vertex1", 0);
  cgGLLoadProgram(cg_vertex_program_1);
  cg_vertex_program_1t = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_vertex_profile, "vertex1t", 0);
  cgGLLoadProgram(cg_vertex_program_1t);
  cg_vertex_program_2 = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_vertex_profile, "vertex2", 0);
  cgGLLoadProgram(cg_vertex_program_2);
  cg_vertex_program_2r = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_vertex_profile, "vertex2r", 0);
  cgGLLoadProgram(cg_vertex_program_2r);
  cg_vertex_program_2t = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_vertex_profile, "vertex2t", 0);
  cgGLLoadProgram(cg_vertex_program_2t);

  // create fragment profile
  cg_fragment_profile = CG_PROFILE_FP30;
  cgGLSetOptimalOptions(cg_fragment_profile);

  // load fragment programs
  cg_fragment_program_1 = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_fragment_profile, "fragment1", 0);
  cgGLLoadProgram(cg_fragment_program_1);
  cg_fragment_program_2 = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_fragment_profile, "fragment2", 0);
  cgGLLoadProgram(cg_fragment_program_2);
  cg_fragment_program_2r = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_fragment_profile, "fragment2r", 0);
  cgGLLoadProgram(cg_fragment_program_2r);
  cg_fragment_program_3 = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_fragment_profile, "fragment3", 0);
  cgGLLoadProgram(cg_fragment_program_3);
  cg_fragment_program_3t = cgCreateProgramFromFile(cg_context, CG_SOURCE, CG_PROGRAM_FILENAME, cg_fragment_profile, "fragment3t", 0);
  cgGLLoadProgram(cg_fragment_program_3t);



  ///////////////////////////
  // OpenGL Initialization //
  //////////////////////////

  // generate texture names
  glGenTextures(1, &gl_gauss_texture);
  glGenTextures(1, &gl_framebuffer_texture);
  glGenTextures(8, gl_framebuffer_texturet);

  // Gauss texture
  generateGaussTexture();
}

void Renderer::points(GLenum format, const GLvoid* vertices) {
  glInterleavedArrays(format, 0, vertices);
}

void Renderer::depthOffset(GLfloat d) {
  depth_offset = d;
}

void Renderer::maxOverlap(GLint o) {
  max_overlap = o;
  generateGaussTexture();
}

void Renderer::cutoffRadius(GLfloat r) {
  cutoff_radius = r;
  generateGaussTexture();
}

void Renderer::splatRadius(GLfloat r) {
  splat_radius = r;
}


void Renderer::render(GLint first, GLsizei num) {
  ////////////////////
  // Initialization //
  ////////////////////

  // Variable declarations
  CGparameter cg_focus, cg_depth_offset, cg_gauss, cg_framebuffer_texture, cg_cutoff_radius, cg_scale_rotate;
  GLint gl_viewport[4];
  GLfloat gl_projection_matrix[16];
  GLfloat gl_focal_length[2];

  // Get camera calibration
  glGetIntegerv(GL_VIEWPORT, gl_viewport);
  glGetFloatv(GL_PROJECTION_MATRIX, gl_projection_matrix);
  gl_focal_length[0] = gl_projection_matrix[0]/2.0*gl_viewport[2];
  gl_focal_length[1] = gl_projection_matrix[5]/2.0*gl_viewport[3];
    
  // enable Cg profiles
  cgGLEnableProfile(cg_vertex_profile);
  cgGLEnableProfile(cg_fragment_profile);

  // allow setting point size by vertex program
  glEnable(GL_VERTEX_PROGRAM_POINT_SIZE_ARB);

  // enable point sprites
  glTexEnvf(GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_DECAL);
  glEnable(GL_POINT_SPRITE_NV);

  // enable depth test
  glEnable(GL_DEPTH_TEST);

  // enable gauss texture
  glEnable(GL_TEXTURE_1D);
  glBindTexture(GL_TEXTURE_1D, gl_gauss_texture);

  // Get Scaling and Rotation part of modelview matrix
  GLfloat mv_scale_rotate[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, mv_scale_rotate);
  mv_scale_rotate[3] = mv_scale_rotate[4];
  mv_scale_rotate[4] = mv_scale_rotate[5];
  mv_scale_rotate[5] = mv_scale_rotate[6];
  mv_scale_rotate[6] = mv_scale_rotate[8];
  mv_scale_rotate[7] = mv_scale_rotate[9];
  mv_scale_rotate[8] = mv_scale_rotate[10];


  ////////////
  // Pass 1 //
  ////////////

  // bind programs
  cgGLBindProgram(cg_vertex_program_1);
  cgGLBindProgram(cg_fragment_program_1);

  // bind and set parameters
  cg_focus = cgGetNamedParameter(cg_vertex_program_1, "focus");
  cg_depth_offset = cgGetNamedParameter(cg_vertex_program_1, "depth_offset");
  cg_cutoff_radius = cgGetNamedParameter(cg_vertex_program_1, "cutoff_radius");
  cg_scale_rotate = cgGetNamedParameter(cg_vertex_program_1, "scale_rotate");
  cgGLSetParameter2f(cg_focus, gl_projection_matrix[0]/2.0*gl_viewport[2], gl_projection_matrix[5]/2.0*gl_viewport[3]);
  cgGLSetParameter1f(cg_depth_offset, depth_offset);
  cgGLSetParameter1f(cg_cutoff_radius, splat_radius);
  cgGLSetMatrixParameterfc(cg_scale_rotate, mv_scale_rotate);
  
  // write-protect color buffer and write-enable depth buffer
//  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glDepthMask(GL_TRUE);

  // draw vertices //
  glDrawArrays(GL_POINTS, first, num);



  ////////////
  // Pass 2 //
  ////////////

  // enable alpha blending
  glEnable(GL_BLEND);
  glBlendEquation(GL_FUNC_ADD);
  glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE, GL_ONE, GL_ONE);

  // bind programs
  cgGLBindProgram(cg_vertex_program_2);
  cgGLBindProgram(cg_fragment_program_2);

  // bind and set parameters
  cg_focus = cgGetNamedParameter(cg_vertex_program_2, "focus");
  cg_cutoff_radius = cgGetNamedParameter(cg_vertex_program_2, "cutoff_radius");
  cg_gauss = cgGetNamedParameter(cg_fragment_program_2, "gauss");

  cg_scale_rotate = cgGetNamedParameter(cg_vertex_program_2, "scale_rotate");
  cgGLSetParameter2f(cg_focus, gl_projection_matrix[0]/2.0*gl_viewport[2], gl_projection_matrix[5]/2.0*gl_viewport[3]);
  cgGLSetParameter1f(cg_cutoff_radius, splat_radius);
  cgGLSetTextureParameter(cg_gauss, gl_gauss_texture);
  cgGLEnableTextureParameter(cg_gauss);
  cgGLSetMatrixParameterfc(cg_scale_rotate, mv_scale_rotate);
 
  // write-enable color buffer and write-protect depth buffer
//   glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glDepthMask(GL_FALSE);

  // draw vertices //
  glDrawArrays(GL_POINTS, first, num);

  /////////////////////////
  // Reset OpenGL states //
  /////////////////////////

  // disable textures
  glDisable(GL_TEXTURE_1D);
  cgGLDisableTextureParameter(cg_gauss);

  // enable writing into depth buffer
  glDepthMask(GL_TRUE);

  // disable Cg profiles
  cgGLDisableProfile(cg_vertex_profile);
  cgGLDisableProfile(cg_fragment_profile);



  ////////////
  // Pass 3 //
  ////////////

  cgGLEnableProfile(cg_fragment_profile);
  cgGLBindProgram(cg_fragment_program_3);

  glEnable(GL_TEXTURE_RECTANGLE_NV);
  glBindTexture(GL_TEXTURE_RECTANGLE_NV, gl_framebuffer_texture);
  glCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGBA, 0, 0, gl_viewport[2], gl_viewport[3], 0);
  
  cg_framebuffer_texture = cgGetNamedParameter(cg_fragment_program_3, "texture");
  cgGLSetTextureParameter(cg_framebuffer_texture, gl_framebuffer_texture);
  cgGLEnableTextureParameter(cg_framebuffer_texture);

  glDisable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glColor4f(1,0,0,1);
  glBegin(GL_QUADS);

  glTexCoord2f(gl_viewport[2],gl_viewport[3]);
  glVertex3f(1,1,0);
  glTexCoord2f(0,gl_viewport[3]);
  glVertex3f(-1,1,0);
  glTexCoord2f(0,0);
  glVertex3f(-1,-1,0);
  glTexCoord2f(gl_viewport[2],0);
  glVertex3f(1,-1,0);
  glEnd();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  glDisable(GL_TEXTURE_RECTANGLE_NV);

  cgGLDisableTextureParameter(cg_framebuffer_texture);
  cgGLDisableProfile(cg_fragment_profile);
}





// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
