#ifndef _RENDERER_H
#define _RENDERER_H

#ifdef _WIN32
#  include <windows.h>
#  include <GL/gl.h>
#  include <GL/glext.h>
#  include "gl_extensions.h"
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/gl.h>
#  include <GL/glext.h>
#endif

#include <Cg/cg.h>
#include <Cg/cgGL.h>


#ifndef _EWAGLRENDERERWIDGET_H_
#define _EWAGLRENDERERWIDGET_H_


#include "../../../DataTypes/src/MyDataTypes.h"
#include "../../../Core/Interfaces/src/RendererInterface.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "EWAGLRendererConfiguration.h"

#include <qgl.h>
#include <qevent.h>
#include <vector>

#include "../../../Core/CoreTools/src/CoreTools.h"

struct MySurfel {
	float texture[2];
	float color[4];
	float normal[3];
	float position[3];
};


class Renderer {
public:
	Renderer();
	void init();
	void points(GLenum format, const GLvoid* vertices);
	void depthOffset(GLfloat d);
	void maxOverlap(GLint o);
	void cutoffRadius(GLfloat r);
	void splatRadius(GLfloat r);
	void render(GLint first, GLsizei num);

private:
	CGcontext cg_context;
	CGprofile cg_vertex_profile, cg_fragment_profile;
	CGprogram cg_vertex_program_1, cg_vertex_program_1t, cg_fragment_program_1, cg_vertex_program_2, cg_vertex_program_2t, cg_vertex_program_2r, cg_fragment_program_2, cg_fragment_program_2r, cg_fragment_program_3t, cg_fragment_program_3;

	GLuint gl_gauss_texture, gl_framebuffer_texture, gl_framebuffer_texturet[8];

	GLfloat cutoff_radius;
	GLfloat splat_radius;
	GLint max_overlap;
	GLfloat depth_offset;

	void generateGaussTexture();
};

#endif






class EWAGLRendererWidget : public QGLWidget
{
	Q_OBJECT

public:
	friend class Renderer;

private:
	Renderer renderer;

public:

	EWAGLRendererWidget (const bool isVisible, EWAGLRendererConfiguration *rendererConfiguration, QWidget *parent = 0, const char *name = 0);
	virtual ~EWAGLRendererWidget();

	void setVisible (const bool enable);
	bool isVisible() const;

	QSize sizeHint() const;
	QSize minimumSizeHint() const;
	QSizePolicy sizePolicy() const;

	/**
	 * Enables/disables the lighting in EWAGL. The
	 * image is <em>not</em> re-rendered automatically.
	 *
	 * @param enable
	 *        set to <code>true</code> if lighting should be enabled; <code>false</code> else
	 */
	void setShadingEnabled (const bool enable);

	/**
	 * Returns the status of the lighting in EWAGL.
	 *
	 * @return <code>true</code> if lighting in EWAGL is enabled; <code>false</code> else
	 */
	bool isShadingEnabled() const;

	void setLightDirection (const Vector3D newLightDirection);

	Vector3D getLightDirection() const;

	void setBackgroundColor(QRgb newBackgroundColor);
	QRgb getBackgroundColor() const;

	/**
	 * Sets the final projection matrix of the <code>Scene</code> which will be used for the next rendering operation.
	 * This is normally the matrix as returned by <code>Scene::getTransformationMatrix</code>, but any transformation
	 * matrix can be used here.
	 * 
	 * @param newSceneViewMatrix
	 *        the <code>TransformationMatrix16f</code> which specifies the projection matrix of the
	 *        <code>Scene</code>
	 * @see Scene#getTransformationMatrix
	 * @see #renderFrame
	 */
	void setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix);

	/**
	 * Returns the projection matrix of the <code>Scene</code> which will be used for the next rendering operation.
	 *
	 * @param sceneViewMatrix
	 *        a <code>TransformationMatrix16f</code> where the projection matrix of the <code>Scene</code>
	 *        will be stored into
	 */
	void getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const;

	/**
	 * Sets the <code>ViewFrustum</code> which will be used for the next rendering operation.
	 *
	 * @param newViewFrustum
	 *        the <code>ViewFrustum</code> which will be used for the next rendering operation
	 * @see #renderFrame
	 */
	void setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum);

	/**
	 * Returns the <code>viewFrustum</code> which will be used for the next rendering operation.
	 *
	 * @return the <code>ViewFrustum</code> which will be used for the next rendering operation
	 * @see #renderFrame
	 */
	MyDataTypes::ViewFrustum getViewFrustum() const;

	void setViewPortSize (const QSize newViewPortSize);
	QSize getViewPortSize() const;

	/**
	 * Set if the selection should be rendered
	 *
	 * @param showSelection
	 *			true if the selected surfels should be rendered differently
	 */
	void setSelectionVisualizationEnabled(const bool showSelection);

	/**
	 * Returns whether the selected surfels are rendered differently
	 */
	bool isSelectionVisualizationEnabled() const;

	
signals:

	/**
	 * Emitted whenever this <code>EWAGLRendererWidget</code> has repainted its content.
	 *
	 * <b>Note:</b> Connected to the <code>EWAGLRenderer</code> which on its turn emits
	 *              the signal <code>rendererWidgetRepainted</code>.
	 */
	void repainted();

protected:

	void mousePressEvent (QMouseEvent *me);
	void mouseMoveEvent (QMouseEvent *me);
	void mouseReleaseEvent (QMouseEvent *me);
	void keyPressEvent (QKeyEvent *ke);
	void wheelEvent (QWheelEvent *we);
	
	void initializeGL();
	void resizeGL (int width, int height);
	void paintGL();



private:

	
	friend class EWAGLRenderer;

	EWAGLRendererConfiguration          *rendererConfiguration;
	Vector3D                             lightDirection;	
	
	std::vector< std::vector<MySurfel> >		 vertexArrays;
	std::vector<float>                   averageRadii;
	                                                                    // its own vertex array
	MyDataTypes::TransformationMatrix16f sceneViewMatrix;
	MyDataTypes::ViewFrustum             viewFrustum;
	bool                                 shadingEnabled,
		                                 updateViewPortSize,            // true whenever the view port size
	                                                                    // in the Configuration should be updated
										 visible;

	QColor								 backgroundColor;

	bool								 bShowSelection;				// true if selected surfels should be rendered differently

    bool                                 firstCall;
	
	// updates the EWAGL light direction according to 'lightDirection'
	// updateScene: true, if scene should be re-rendered with the new light direction
	void updateLightDirection (const bool updateScene);
	
	inline void showSelection(SurfelInterface *surfel,float& r_comp,float& g_comp,float& b_comp);

    void buildVertexArrays();

	// returns the framebuffer / makes a screenshot
	QImage renderPixmap();


private slots:
	
	void handleObjectModified();
	void handleActiveObjectChanged();
	void handleObjectAdded();
	void handleObjectRemoved();	
	void handleSelectionChanged (SurfelInterface::Flag);
	void handleCutOffRadiusChanged (const float newCutOffRadius);
    void handleGaussDecayRadiusChanged (const float newGaussRadius);

	// connected to the EWAGLRendererConfiguration, signal configurationApplied: re-renders the image
	void handleConfigurationApplied();

private:
    RendererInterface::InteractStateEnum  modelInteractState;
    RendererInterface::InteractStateEnum  viewpointInteractState;
  
public:
    virtual RendererInterface::InteractStateEnum  getModelInteractState(void) const;
    virtual RendererInterface::InteractStateEnum  getViewpointInteractState(void) const;

    virtual void  setInteractState(RendererInterface::InteractStateEnum state,
				                   RendererInterface::InteractStateAspectEnum  target=RendererInterface::INTERACT_MODEL_AND_VIEWPOINT);

};

#endif  // _OPENGLRENDERERWIDGET_H_


inline void EWAGLRendererWidget::showSelection(SurfelInterface *surfel,float& r_comp, float& g_comp, float& b_comp) {
		
	// modify color if surfel is selected and visualization of the selection is enabled
	if(bShowSelection && surfel->isFlagOn (SurfelInterface::EMPHASISE) == true) {
		if (surfel->isFlagOn (SurfelInterface::SELECTED1) == true) {
			r_comp *= 0.5f;
			g_comp *= 0.5f;
			b_comp = b_comp * 0.5f + 122.0f;
		}
		if (surfel->isFlagOn (SurfelInterface::SELECTED2) == true) {
			r_comp = r_comp * 0.5f + 122.0f;
			g_comp *= 0.5f;
			b_comp *= 0.5f;
		}
		if (surfel->isFlagOn (SurfelInterface::SELECTED3) == true) {
			r_comp *= 0.5f;
			g_comp = g_comp * 0.5f + 122.0f;
			b_comp *= 0.5f;
		}
	}
}













 /* Some Emacs-Hints -- please don't remove:
 *
 *  Local Variables:
 *  mode:C++
 *  tab-width:4
 *  comment-column:72
 *  End:
 */

